@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1993-2006 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
  */
 
 /* block.cc : block operations */
-/* $Id: block.cc,v 1.22 2001/04/09 08:24:56 lav Exp $ */
+/* $Id: block.cc,v 1.28 2006/04/12 05:46:04 lav Exp $ */
 
 #include <config.h>
 #include <assert.h>
@@ -83,6 +83,14 @@ char   CharAtLC(num l,num c)
    Last=TextPoint(l,c);
    return((EolAt(Last)||Last.Col()!=c||Last.Line()!=l)?' ':CharAt(Last));
 }
+#if USE_MULTIBYTE_CHARS
+wchar_t WCharAtLC(num l,num c)
+{
+   static TextPoint Last;
+   Last=TextPoint(l,c);
+   return((EolAt(Last)||Last.Col()!=c||Last.Line()!=l)?' ':WCharAt(Last));
+}
+#endif
 void   NewLine()
 {
    InsertBlock(EolStr,EolSize);
@@ -204,12 +212,11 @@ int   RDelete()
       else
       {
 	 offs o=CurrentPos;
-	 while(j<oldcol2 && !Eol())
+	 while(GetCol()<oldcol2 && !Eol())
 	 {
 	    if(Char()=='\t')
 	       ExpandTab();
-	    MoveRight();
-	    j++;
+	    MoveRightOverEOL();
 	 }
 	 DeleteBlock(CurrentPos-o,0);
       }
@@ -584,7 +591,7 @@ void    DoIndent(int i)
       for(li=BlockBegin.Line(); li<=BlockEnd.Line(); li++)
       {
          HardMove(li,BlockBegin.Col());
-         while(!Eol() && isspace(Char()))
+         while(!Eol() && Space())
          {
             ExpandTab();
             MoveRight();
@@ -602,7 +609,7 @@ void    DoIndent(int i)
    do
    {
       space=0;
-      while(!Eol() && isspace(Char()))
+      while(!Eol() && Space())
       {
          if(Char()=='\t')
             space=Tabulate(space);
@@ -637,13 +644,13 @@ void    DoUnindent(int i)
       for(li=BlockBegin.Line(); li<=BlockEnd.Line(); li++)
       {
          HardMove(li,BlockBegin.Col());
-         while(!Eol() && isspace(Char()))
+         while(!Eol() && Space())
          {
             ExpandTab();
             MoveRight();
          }
          HardMove(li,BlockBegin.Col());
-         for(j=i; j>0 && !Eol() && isspace(Char()); j--)
+         for(j=i; j>0 && !Eol() && Space(); j--)
             DeleteChar();
       }
       return;
@@ -655,7 +662,7 @@ void    DoUnindent(int i)
    do
    {
       space=0;
-      while(!Eol() && isspace(Char()))
+      while(!Eol() && Space())
       {
          if(Char()=='\t')
             space=Tabulate(space);
@@ -928,27 +935,117 @@ void  Transform(byte (*func)(byte))
    CurrentPos=oldpos;
    stdcol=GetCol();
 }
+#if USE_MULTIBYTE_CHARS
+wctrans_t trans_toupper;
+wchar_t ToupperW(wchar_t wc)
+{
+   return towctrans(wc,trans_toupper);
+}
+wctrans_t trans_tolower;
+wchar_t TolowerW(wchar_t wc)
+{
+   return towctrans(wc,trans_tolower);
+}
+wchar_t InverseW(wchar_t wc)
+{
+   if(iswupper(wc))
+      return towctrans(wc,trans_tolower);
+   else
+      return towctrans(wc,trans_toupper);
+}
+void  TransformW(wchar_t (*func)(wchar_t))
+{
+   TextPoint oldpos=CurrentPos;
+
+   if(hide || (!rblock && BlockBegin==BlockEnd))
+   {
+      while(!Eol())
+         ReplaceWCharMove(func(WChar()));
+      flag|=REDISPLAY_LINE;
+   }
+   else
+   {
+      if(rblock)
+      {
+         num i;
+         num line1=BlockBegin.Line();
+         num line2=BlockEnd.Line();
+         num col1 =BlockBegin.Col();
+         num col2 =BlockEnd.Col();
+
+         for(i=line1; i<=line2; i++)
+         {
+	    HardMove(i,col1);
+            if(col2>col1)
+            {
+	       while(GetCol()<col2 && !Eol())
+                  ReplaceWCharMove(func(WChar()));
+            }
+            else
+            {
+               while(!Eol())
+                  ReplaceWCharMove(func(WChar()));
+	    }
+         }
+      }
+      else
+      {
+         CurrentPos=BlockBegin;
+         while(CurrentPos<BlockEnd)
+	    ReplaceWCharMove(func(WChar()));
+      }
+      flag=REDISPLAY_ALL;
+   }
+   CurrentPos=oldpos;
+   stdcol=GetCol();
+}
+#endif
 
 void    ConvertToUpper()
 {
    if(DragMark)
       UserStopDragMark();
 
-   Transform(Toupper);
+#if USE_MULTIBYTE_CHARS
+   if(mb_mode)
+   {
+      trans_toupper=wctrans("toupper");
+      TransformW(ToupperW);
+   }
+   else
+#endif
+      Transform(Toupper);
 }
 void    ConvertToLower()
 {
    if(DragMark)
       UserStopDragMark();
 
-   Transform(Tolower);
+#if USE_MULTIBYTE_CHARS
+   if(mb_mode)
+   {
+      trans_tolower=wctrans("tolower");
+      TransformW(TolowerW);
+   }
+   else
+#endif
+      Transform(Tolower);
 }
 void    ExchangeCases()
 {
    if(DragMark)
       UserStopDragMark();
 
-   Transform(Inverse);
+#if USE_MULTIBYTE_CHARS
+   if(mb_mode)
+   {
+      trans_toupper=wctrans("toupper");
+      trans_tolower=wctrans("tolower");
+      TransformW(InverseW);
+   }
+   else
+#endif
+      Transform(Inverse);
 }
 void    BlockType()
 {
