@@ -56,7 +56,7 @@ void  init_attrs();
 extern attr attr_table[MAX_COLOR_NO];
 
 static inline
-attr *find_attr(int no)
+const attr *find_attr(int no)
 {
    return attr_table+no;
 }
@@ -64,13 +64,14 @@ attr *find_attr(int no)
 extern char color_descriptions[MAX_COLOR_NO*2][256];
 
 void  ParseColors();
-color *FindColor(color *pal,int no);
-void  DescribeColors(color *,color *);
+const color *FindColor(const color *pal,int no);
+static inline color *FindColor(color *pal,int no) { return const_cast<color*>(FindColor(const_cast<const color*>(pal),no)); }
+void  DescribeColors(const color *,const color *);
 void  DumpDefaultColors(FILE *);
 
 extern color color_pal[MAX_COLOR_NO+1];
 extern color bw_pal[MAX_COLOR_NO+1];
-extern color default_color_pal[MAX_COLOR_NO+1];
-extern color default_bw_pal[MAX_COLOR_NO+1];
+extern const color default_color_pal[MAX_COLOR_NO+1];
+extern const color default_bw_pal[MAX_COLOR_NO+1];
 
 #define NO_COLOR (-1)
