@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2000 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1993-2006 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: edit.cc,v 1.26 2001/06/08 10:31:00 lav Exp $ */
+/* $Id: edit.cc,v 1.42 2006/11/29 07:36:34 lav Exp $ */
 
 /* edit.c : main editor loop */
 
@@ -33,6 +33,9 @@
 #include <locale.h>
 #include <time.h>
 #include <errno.h>
+#ifdef HAVE_LANGINFO_H
+# include <langinfo.h>
+#endif
 #include "edit.h"
 #include "calc.h"
 #include "keymap.h"
@@ -41,6 +44,7 @@
 #include "options.h"
 #include "about.h"
 #include "search.h"
+#include "undo.h"
 #ifdef WITH_MOUSE
 # include "mouse.h"
 #endif
@@ -83,7 +87,8 @@ void  GoToLineNum(num line_num)
    stdcol=GetCol();
 }
 
-int   getcode()
+History CodeHistory;
+long getcode(const char *prompt)
 {
    long  i;
    static char ch[10];
@@ -94,17 +99,31 @@ int   getcode()
 
    getcode_active=true;
 
-   if(getstring("Char: ",ch,sizeof(ch)-1)<1)
+   if(getstring(prompt,ch,sizeof(ch)-1,&CodeHistory)<1)
    {
       getcode_active=false;
       return(-1);
    }
    getcode_active=false;
    i=strtol(ch,0,0);
-   if(i<0 || i>255)
-      return(-1);
    return((int)i);
 }
+int getcode_char()
+{
+   long ch=getcode("Char: ");
+   if(ch<0 || ch>=256)
+      return -1;
+   return (int)ch;
+}
+#if USE_MULTIBYTE_CHARS
+wchar_t getcode_wchar()
+{
+   long ch=getcode("Wide Char: ");
+   if(ch<0)
+      return -1;
+   return (wchar_t)ch;
+}
+#endif
 
 void ProcessDragMark()
 {
@@ -150,7 +169,6 @@ void    Edit()
 
       ClearMessage();
       SyncTextWin();
-      StatusLine();
       SetCursor();
 
       action=GetNextAction();
@@ -191,7 +209,9 @@ void    Edit()
       proc=GetActionProc(EditorActionProcTable,action);
       if(proc)
       {
+	 undo.BeginUndoGroup();
 	 proc();
+	 undo.EndUndoGroup();
 	 continue;
       }
       if(action!=NO_ACTION)
@@ -223,6 +243,7 @@ void    Edit()
 	       continue;
 	    }
 	 }
+	 undo.BeginUndoGroup();
 	 if(((insert && !right) || Eof())
 	 && !buffer_mmapped)
 	 {
@@ -241,6 +262,7 @@ void    Edit()
 	       right=1;
 	 }
 	 flag|=REDISPLAY_LINE;
+	 undo.EndUndoGroup();
       }
       else
       {
@@ -275,6 +297,7 @@ void    Edit()
 	    }
 	    continue;
 	 }
+	 undo.BeginUndoGroup();
 	 switch(key)
 	 {
 	 case('\n'):
@@ -285,16 +308,12 @@ void    Edit()
 	    break;
 	 default:   /* not a newline and not a tab */
 	    if(insert || Eol() || (Char()=='\t' && Tabulate(GetCol())!=(GetCol()+1)))
-	    {
 	       UserInsertChar(key);
-	    }
 	    else
-	    {
-	       InsertChar(key);
-	       DeleteChar();
-	    }
+	       UserReplaceChar(key);
 	    flag|=REDISPLAY_LINE;
 	 }
+	 undo.EndUndoGroup();
       }
    }
 }
@@ -404,7 +423,8 @@ void    Initialize()
 {
    FILE    *f;
 
-   InitTables();
+   InitModifyKeyTables();
+   init_chset();
    initcalc();
 
 #ifndef MSDOS
@@ -413,7 +433,7 @@ void    Initialize()
    mkdir(filename,0700);
    strcat(filename,"/tmp");
    mkdir(filename,0700);
-   sprintf(HstName,"%s/.le/history",HOME);
+   sprintf(HstName,"%s/.le/history2",HOME);
 #else
    sprintf(HstName,"%s/le.hst",HOME);
 #endif
@@ -439,9 +459,11 @@ void    Initialize()
    f=fopen(HstName,"rb");
    if(f && fcntl(fileno(f),F_SETLKW,&l)!=-1)
    {
+      PositionHistory.ReadFrom(f);
       LoadHistory.ReadFrom(f);
       SearchHistory.ReadFrom(f);
-      PositionHistory.ReadFrom(f);
+      ShellHistory.ReadFrom(f);
+      PipeHistory.ReadFrom(f);
       fclose(f);
    }
 
@@ -473,23 +495,35 @@ void    Terminate()
          {
             f=fdopen(fd,"r+b");
 
+            InodeHistory oldPositionHistory;
             History  oldLoadHistory;
             History  oldSearchHistory;
-            InodeHistory oldPositionHistory;
+            History  oldShellHistory;
+            History  oldPipeHistory;
 
+            oldPositionHistory.ReadFrom(f);
             oldLoadHistory.ReadFrom(f);
             oldSearchHistory.ReadFrom(f);
-            oldPositionHistory.ReadFrom(f);
+            oldShellHistory.ReadFrom(f);
+            oldPipeHistory.ReadFrom(f);
+            PositionHistory.Merge(oldPositionHistory);
             LoadHistory.Merge(oldLoadHistory);
             SearchHistory.Merge(oldSearchHistory);
-            PositionHistory.Merge(oldPositionHistory);
+            ShellHistory.Merge(oldShellHistory);
+            PipeHistory.Merge(oldPipeHistory);
 
             rewind(f);
 
+            PositionHistory.WriteTo(f);
             LoadHistory.WriteTo(f);
             SearchHistory.WriteTo(f);
-            PositionHistory.WriteTo(f);
+            ShellHistory.WriteTo(f);
+            PipeHistory.WriteTo(f);
 
+#ifdef HAVE_FTRUNCATE
+	    fflush(f);
+	    ftruncate(fd,ftell(f));
+#endif
             fclose(f);
          }
          close(fd);
@@ -517,8 +551,8 @@ void  PrintUsage(int arg)
 	  "    --dump-colors  dump default color map to stdout and exit\n"
 	  "    --mmap         load file using mmap (read only)\n"
 	  "    --mmap-rw      mmap read-write. Use with extreme caution,\n"
-	  "                   expecially on your hard disk! No undo, all\n"
-	  "                   changes go directly to file/disk.\n"
+	  "                   expecially on your hard disk!\n"
+	  "                   All changes go directly to file/disk.\n"
 	  "    --help         this description\n"
 	  "    --version      print LE version\n"
 	  "\n"
@@ -527,10 +561,21 @@ void  PrintUsage(int arg)
    exit(1);
 }
 
+#if USE_MULTIBYTE_CHARS
+static bool has_widechars()
+{
+   for(int c=' '; c<256; c++)
+      if(btowc(c)>=256)
+	 return 1;
+   return 0;
+}
+#endif
+
 int     main(int argc,char **argv)
 {
    int   optView=-1,opteditmode=-1,optWarpLine=0;
    int	 opt_use_mmap=-1;
+   int	 opt_mb_mode=-1;
    int   opt;
 
    enum {
@@ -540,7 +585,12 @@ int     main(int argc,char **argv)
       PRINT_VERSION,
       CONFIG_FILE,
       USE_MMAP,
-      USE_MMAP_RW
+      USE_MMAP_RW,
+#if USE_MULTIBYTE_CHARS
+      MULTIBYTE,
+      NO_MULTIBYTE,
+#endif
+      MAX_OPTION
    };
 
    static struct option le_options[]=
@@ -554,6 +604,10 @@ int     main(int argc,char **argv)
       {"black-white",no_argument,0,'b'},
       {"color",no_argument,0,'c'},
       {"config",required_argument,0,CONFIG_FILE},
+#if USE_MULTIBYTE_CHARS
+      {"multibyte",no_argument,0,MULTIBYTE},
+      {"no-multibyte",no_argument,0,NO_MULTIBYTE},
+#endif
 #ifdef HAVE_MMAP
       {"mmap",no_argument,0,USE_MMAP},
       {"mmap-rw",no_argument,0,USE_MMAP_RW},
@@ -561,8 +615,8 @@ int     main(int argc,char **argv)
       {0,0,0,0}
    };
 
-   HistoryLine    *hl=0;
    char  newname[256];
+   newname[0]=0;
 
    strcpy(Program,le_basename(argv[0]));
 
@@ -580,6 +634,15 @@ int     main(int argc,char **argv)
     _fmode=O_BINARY;
 #else
    setlocale(LC_ALL,"");
+#if USE_MULTIBYTE_CHARS
+   if(has_widechars())
+      mb_mode=true;
+# if defined(HAVE_NL_LANGINFO) && defined(CODESET)
+   char *cs=nl_langinfo(CODESET);
+   if(cs && !strcasecmp(cs,"UTF-8"))
+      mb_mode=true;
+# endif
+#endif
 
    HOME=getenv("HOME");
    if(HOME==NULL)
@@ -649,6 +712,14 @@ int     main(int argc,char **argv)
 	 ExplicitInitName=true;
 	 strncpy(InitName,optarg,sizeof(InitName)-1);
 	 break;
+#if USE_MULTIBYTE_CHARS
+      case MULTIBYTE:
+	 opt_mb_mode=true;
+	 break;
+      case NO_MULTIBYTE:
+	 opt_mb_mode=false;
+	 break;
+#endif
       }
    }
    if(optUseColor!=-1)
@@ -664,6 +735,10 @@ int     main(int argc,char **argv)
       UseColor=optUseColor;
    if(opt_use_mmap!=-1)
       buffer_mmapped=opt_use_mmap;
+#if USE_MULTIBYTE_CHARS
+   if(opt_mb_mode!=-1)
+      mb_mode=opt_mb_mode;
+#endif
 
    if(optind<argc-1 && argv[optind][0]=='+' && isdigit(argv[optind][1]))
    {
@@ -673,6 +748,7 @@ int     main(int argc,char **argv)
 
    if(optind>=argc)
    {
+      const HistoryLine *hl=0;
       LoadHistory.Open();
       bool first=true;
       for(;;)
@@ -680,10 +756,11 @@ int     main(int argc,char **argv)
 	 hl=LoadHistory.Prev();
 	 if(!hl)
 	    break;
-	 if(hl->line[0] && (first || hl->line[0]!='/')
-	 && access(hl->line,R_OK)!=-1)
+	 const char *f=hl->get_line();
+	 if(*f && (first || f[0]!='/')
+	 && access(f,R_OK)!=-1)
 	 {
-	    strcpy(newname,hl->line);
+	    strcpy(newname,f);
       	    break;
 	 }
 	 first=false;
