@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-2000 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1993-2004 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: highli.cc,v 1.17 2002/08/06 16:01:41 lav Exp $ */
+/* $Id: highli.cc,v 1.19 2006/11/29 08:14:47 lav Exp $ */
 
 #include <config.h>
 
@@ -215,7 +215,7 @@ void InitHighlight()
 	    if(s[0]=='/')
 	    {
 	       // it is a regex for file contents
-	       if(strlen(s)<len)
+	       if(strlen(s)+(s-str)<len)
 		  s[strlen(s)]='|';
 
 	       s++;
