@@ -26,6 +26,7 @@ extern "C" {
    #include <regex.h>
 #else
    #include <rx.h>
+   #define RE_TRANSLATE_TYPE unsigned char *
 #endif
 }
 
@@ -50,7 +51,7 @@ struct syntax_hl
 
    syntax_hl(int color,int mask);
    ~syntax_hl();
-   const char *set_rexp(const char *rexp);
+   const char *set_rexp(const char *rexp,bool icase);
 };
 
 #endif//HIGHLI_H
