@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1993-2004 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: kern.cc,v 1.16 2001/04/13 08:56:01 lav Exp $ */
+/* $Id: kern.cc,v 1.30 2006/06/20 12:06:03 lav Exp $ */
 
 #include <config.h>
 #include   <stdio.h>
@@ -35,6 +35,8 @@
 #include   <assert.h>
 #include   "keymap.h"
 #include   "edit.h"
+#include   "mb.h"
+#include   "undo.h"
 
 #ifndef O_NDELAY
 #define O_NDELAY 0
@@ -55,7 +57,7 @@ bool      newfile=false;
 
 int       DosEol=0;
 int       EolSize=1;
-char      *EolStr="\n";
+const char *EolStr="\n";
 
 InodeInfo   FileInfo;
 InodeHistory PositionHistory;
@@ -177,18 +179,31 @@ void   ToLineEnd()
 
 void   MoveLeftOverEOL()
 {
-   if(Bol() && !hex)
+   if(hex)
+   {
+      CurrentPos-=1;
+      return;
+   }
+   if(Bol())
       CurrentPos-=EolSize;
    else
-      CurrentPos-=1;
+   {
+      MBCheckLeft();
+      CurrentPos-=MBCharSize;
+   }
 }
 
 void   MoveRightOverEOL()
 {
-   if(Eol() && !hex)
+   if(hex)
+   {
+      CurrentPos+=1;
+      return;
+   }
+   if(Eol())
       CurrentPos+=EolSize;
    else
-      CurrentPos+=1;
+      CurrentPos+=CharSize();
 }
 
 void   MoveUp()
@@ -315,15 +330,15 @@ void  CalculateLineCol(num *line,num *col,offs source,offs target)
    }
 }
 
-int   InsertBlock(char *block_left,register num size_left,char *block_right,num size_right)
+int   InsertBlock(const char *block_left,num size_left,const char *block_right,num size_right)
 {
    if(buffer_mmapped)
    {
       return ERR;
    }
 
-   register num   i;
-   register offs  oldoffset;
+   num   i;
+   offs  oldoffset;
    num   num_of_lines,num_of_columns,oldline,oldcol;
    num	 num_of_lines_curr,num_of_columns_curr;
    int   break_at;
@@ -337,7 +352,6 @@ int   InsertBlock(char *block_left,register num size_left,char *block_right,num
       return(OK);
 
    PreModify();
-
    if(size_left>0)
    {
       if(oldptr1>ptr1)
@@ -357,6 +371,9 @@ int   InsertBlock(char *block_left,register num size_left,char *block_right,num
    else
       new_oldmodified=oldmodified;
 
+   if(undo.Enabled())
+      undo.AddChange(new Undo::Insert(block_left,size_left,block_right,size_right));
+
    memmove(buffer+ptr1,block_left,size_left);
    memmove(buffer+ptr2-size_right,block_right,size_right);
 
@@ -392,7 +409,7 @@ int   InsertBlock(char *block_left,register num size_left,char *block_right,num
          break;
       }
    }
-   for(register TextPoint *scan=TextPoint::base; scan; scan=scan->next)
+   for(TextPoint *scan=TextPoint::base; scan; scan=scan->next)
    {
       if(scan->offset>oldoffset || scan==&TextEnd)
       {
@@ -410,10 +427,12 @@ int   InsertBlock(char *block_left,register num size_left,char *block_right,num
       }
       else if(scan==&CurrentPos)
       {
-         scan->col=num_of_columns_curr;
-         scan->line+=num_of_lines_curr;
+	 if(!(scan->flags&LINEUNDEFINED))
+	 {
+	    scan->line+=num_of_lines_curr;
+	    scan->flags|=COLUNDEFINED;
+	 }
          scan->offset+=size_left;
-	 scan->flags&=~(COLUNDEFINED|LINEUNDEFINED);
       }
    }
 
@@ -698,10 +717,12 @@ int   DeleteBlock(num left,num right)
       oldptr1=ptr1;
    if(oldptr2>ptr2 && right>0)
       oldptr2=ptr2;
-
    if(oldptr1==ptr1 && oldptr2==ptr2)
       oldmodified=modified;
 
+   if(undo.Enabled())
+      undo.AddChange(new Undo::Delete(buffer+ptr1-left,left,buffer+ptr2,right));
+
    size=left+right;
    base=Offset()-left;
 
@@ -760,9 +781,9 @@ int   DeleteBlock(num left,num right)
    return(OK);
 }
 
-int   ReplaceBlock(char *block,num size)
+int   ReplaceBlock(const char *block,num size)
 {
-   if(!buffer_mmapped)
+   if(!buffer_mmapped && !undo.Enabled() && !undo.Locked())
    {
       int res=InsertBlock(block,size);
       if(res==OK)
@@ -776,12 +797,23 @@ int   ReplaceBlock(char *block,num size)
    offs base=Offset();
    offs newline=GetLine();
 
-   if(base>Size()-size)
+   num oldsize=size;
+   if(base>Size()-oldsize)
+      oldsize=Size()-base;
+
+   if(buffer_mmapped)
    {
-      size=Size()-base;
+      if(size>oldsize)
+	 size=oldsize;
       if(size<=0)
 	 return ERR;
    }
+   else
+   {
+      if(size>oldsize && GetSpace(size-oldsize)!=OK)
+	 return ERR;
+   }
+   PreModify();
 
    int break_at=-1;
    int i;
@@ -794,13 +826,21 @@ int   ReplaceBlock(char *block,num size)
       }
    }
 
+   MBCheckLeftAt(base);
+   int mb_size0=MBCharSize;
+
    num num_of_lines=0;
    offs o;
-   for(o=base+1; o<=base+size; o++)
+   for(o=base+1; o<=base+oldsize; o++)
       if(BolAt(o))
 	 num_of_lines++;
 
-   memmove(buffer+base,block,size);
+   if(undo.Enabled())
+      undo.AddChange(new Undo::Replace(buffer+ptr2,oldsize,block,size));
+
+   memmove(buffer+ptr2-(size-oldsize),block,size);
+   oldptr2=ptr2;
+   ptr2-=(size-oldsize);
 
    for(o=base; o<base+size; o++)
       if(EolAt(o))
@@ -816,9 +856,13 @@ int   ReplaceBlock(char *block,num size)
      }
    }
 
+   MBCheckLeftAt(base);
+   int mb_size1=MBCharSize;
+   int mb_size_max=(mb_size0>mb_size1?mb_size0:mb_size1);
+
    for(TextPoint *scan=TextPoint::base; scan; scan=scan->next)
    {
-      if(scan->offset>base+size)
+      if(scan->offset>=base+size)
       {
          if(!(scan->flags&LINEUNDEFINED))
          {
@@ -836,27 +880,24 @@ int   ReplaceBlock(char *block,num size)
 	 scan->offset=base+size;
 	 scan->flags|=COLUNDEFINED|LINEUNDEFINED;
       }
+      else if(scan->offset+mb_size_max>base)
+	 scan->flags|=COLUNDEFINED;
    }
 
+   // when mmapped, changes are committed to disk automatically.
+   if(!buffer_mmapped)
+      modified=1;
+
    return OK;
 }
 
 int   ReplaceCharMove(byte ch)
 {
-   int res;
-   if(!buffer_mmapped)
-   {
-      res=InsertChar(ch);
-      if(res==OK)
-	 DeleteChar();
-   }
-   else
-   {
-      res=ReplaceChar(ch);
-      if(res==OK)
-	 MoveRight();
-   }
-   return res;
+   int res=ReplaceChar(ch);
+   if(res!=OK)
+      return res;
+   MoveRight();
+   return OK;
 }
 
 int   Undelete()
@@ -922,8 +963,31 @@ void   DeleteToEOL()
 
 offs   NextLine(offs ptr)
 {
-   while(!EofAt(ptr) && !BolAt(++ptr));
-   return(ptr);
+   char eol=EolStr[EolSize-1];
+   char *found;
+   while(ptr<ptr1)
+   {
+      found=(char*)memchr(buffer+ptr,eol,ptr1-ptr);
+      if(!found)
+      {
+	 ptr=ptr1;
+	 break;
+      }
+      ptr=found-buffer;
+      if(BolAt(++ptr))
+	 return ptr;
+   }
+   offs size=Size();
+   for(;;)
+   {
+      found=(char*)memchr(buffer+ptr+GapSize,eol,size-ptr);
+      if(!found)
+	 return size;
+      ptr=found-buffer-GapSize;
+      if(BolAt(++ptr))
+	 return ptr;
+   }
+   /*NOTREACHED*/
 }
 offs   NextNLines(offs ptr,num n)
 {
@@ -976,6 +1040,7 @@ void  EmptyText()
       }
    }
    FileName[0]=0;
+   undo.Clear();
    TextPoint::ResetTextPoints();
 
    if(buffer)
@@ -1005,17 +1070,26 @@ void  EmptyText()
    EolStr="\n";
 }
 
+bool IsAlNumAt(num o)
+{
+#if USE_MULTIBYTE_CHARS
+   if(mb_mode)
+      return iswalnum(WCharAt(o));
+#endif
+   return isalnum(CharAt(o)) || isrussian(CharAt(o));
+}
+
 char  *GetWord()
 {
    static   char  word[256];
    num         shift=(-1);
    int         i;
 
-   while(isalnum(CharRel(shift)) || CharRel(shift)=='.' || CharRel(shift)=='_')
+   while(IsAlNumRel(shift) || CharRel(shift)=='.' || CharRel(shift)=='_')
      shift--;
    shift++;
    i=0;
-   while(i<255 && (isalnum(CharRel(shift)) || CharRel(shift)=='.' || CharRel(shift)=='_'))
+   while(i<255 && (IsAlNumRel(shift) || CharRel(shift)=='.' || CharRel(shift)=='_'))
      word[i++]=CharRel(shift++);
    word[i]=0;
    return(word);
@@ -1114,3 +1188,55 @@ offs  ScanForCharForward(offs start,byte ch)
    }
    return -1;
 }
+
+void  InsertAutoindent(num oldcol)
+{
+   int   UseTabsNow=UseTabs;
+   num   cnt;
+   num   oldmargin;
+   num   newmargin=0;
+
+   if(View || !Bol())
+      return;
+
+   offs o=Offset()-EolSize;
+   for(;;)
+   {
+      oldmargin=MarginSizeAt(o);
+      if(BofAt(o) || oldmargin!=-1)
+	 break;
+      o=PrevLine(o);
+   }
+   if(TabsInMargin)
+      UseTabsNow=1;
+
+   if(oldmargin==oldcol)
+      newmargin=oldmargin;
+   else if(oldcol==0)
+      newmargin=0;
+   else if(oldmargin==-1)
+      newmargin=oldcol/IndentSize*IndentSize;
+   else
+      newmargin=oldmargin;
+
+   cnt=newmargin;
+   if(Text && Eol() && !(!UseTabs && UseTabsNow))
+   {
+      stdcol=cnt;
+      return;
+   }
+   if(UseTabsNow)
+   {
+       while(cnt>=TabSize)
+       {
+           cnt-=TabSize;
+           InsertChar('\t');
+       }
+   }
+   while(cnt>0)
+   {
+       cnt--;
+       InsertChar(' ');
+   }
+   stdcol=GetCol();
+}
