@@ -32,14 +32,14 @@
 #include "getch.h"
 #include <term.h>
 
-char  StringTyped[256];
+unsigned char StringTyped[256];
 int   StringTypedLen;
 
 int   FuncKeysNum=12;
 
 int   MouseCounter=0;
 
-ActionNameRec  ActionNameTable[]=
+const ActionNameRec ActionNameTable[]=
 {
    {CHAR_LEFT,"backward-char"},
    {CHAR_RIGHT,"forward-char"},
@@ -92,6 +92,12 @@ ActionNameRec  ActionNameTable[]=
    {COPY_FROM_UP,"copy-from-up"},
    {COPY_FROM_DOWN,"copy-from-down"},
 
+// Undo/redo
+   {UNDO,"undo"},
+   {REDO,"redo"},
+   {UNDO_STEP,"undo-step"},
+   {REDO_STEP,"redo-step"},
+
 // File ops
    {LOAD_FILE,"load-file"},
    {SWITCH_FILE,"switch-file"},
@@ -147,6 +153,8 @@ ActionNameRec  ActionNameTable[]=
    {TABS_EXPAND,"expand-tabs"},
    {TEXT_OPTIMIZE,"optimize-text"},
    {CHOOSE_CHAR,"choose-character"},
+   {CHOOSE_WCHAR,"choose-wide-character"},
+   {CHOOSE_BYTE,"choose-byte"},
    {UNIX_DOS_TRANSFORM,"change-text-type"},
 
 // Options
@@ -158,7 +166,9 @@ ActionNameRec  ActionNameTable[]=
    {SAVE_OPTIONS_LOCAL,"save-options-local"},
 
    {ENTER_CONTROL_CHAR,"quoted-insert"},
-//   {ENTER_CHAR_CODE,UserEnterCharCode},
+   {ENTER_CHAR_CODE,"insert-char-by-code"},
+   {ENTER_WCHAR_CODE,"insert-wchar-by-code"},
+   {ENTER_BYTE_CODE,"insert-byte-by-code"},
 
 //   WINDOW_RESIZE,
 
@@ -202,6 +212,7 @@ ActionNameRec  ActionNameTable[]=
    {LOAD_COLOR_GREEN,"load-color-green"},
    {LOAD_COLOR_WHITE,"load-color-white"},
    {PROGRAMS_OPTIONS,"programs-options"},
+   {UNDO_OPTIONS,"undo-options"},
    {ABOUT,"about"},
    {LOAD_KEYMAP_DEFAULT,"load-keymap-default"},
    {LOAD_KEYMAP_EMACS,  "load-keymap-emacs"},
@@ -255,12 +266,11 @@ struct KeyTreeNode
    struct KeyTreeNode *child;
 };
 
-extern   ActionCodeRec  DefaultActionCodeTable[];
-ActionCodeRec  *ActionCodeTable=DefaultActionCodeTable;
-bool NeedFreeActionCodeTable=false;
+const ActionCodeRec *ActionCodeTable=DefaultActionCodeTable;
+ActionCodeRec *DynamicActionCodeTable;
 //char  *ti_cache[128]={NULL};
 
-char  *GetActionName(int action)
+const char *GetActionName(int action)
 {
    for(int i=0; ActionNameTable[i].action!=-1; i++)
       if(ActionNameTable[i].action==action)
@@ -275,19 +285,19 @@ const char *GetActionCodeText(const char *code)
 
    while(*code)
    {
-      if(iscntrl(*code))
+      unsigned char the_code=*code++;
+      if(iscntrl(the_code))
       {
-         if(*code=='\033')
+         if(the_code=='\033')
             sprintf(store,"\\e");
-         else if(*code>=0 && *code<32)
-            sprintf(store,"^%c",*code+'@');
+         else if(the_code<32)
+            sprintf(store,"^%c",the_code+'@');
          else
-            sprintf(store,"\\%03o",*code);
+            sprintf(store,"\\%03o",the_code);
          store+=strlen(store);
       }
       else
-         *(store++)=*code;
-      code++;
+         *(store++)=the_code;
    }
    *store=0;
    return(code_text);
@@ -330,17 +340,22 @@ static int PrettyCodeScore(const char *c)
 	 if(!(bracket && *c==RIGHT_BRACE))
 	    c--;
 
+	 shift=0;
 	 if(sscanf(term_name,"%1dkf%d",&shift,&fk)==2
 	 || sscanf(term_name,"kf%d",&fk)==1)
 	 {
 	    if(shift)
-	       score+=4*shift;
+	       score+=2+2*shift;
 	    else
 	       score+=2;
+	    if(shift)
+	       sprintf(term_name,"kf%d",shift*FuncKeysNum+fk);
 	 }
 	 else
 	    score+=8;
-
+	 term_str=tigetstr(term_name);
+	 if(term_str==(char*)-1 || !term_str || !*term_str)
+	    return 1000000;
 	 break;
       case('|'):
 	 score+=5;
@@ -516,7 +531,7 @@ KeyTreeNode *AddToKeyTree(KeyTreeNode *curr,int key_code,int delay,int action)
 #define LEFT_BRACE  '{'
 #define RIGHT_BRACE '}'
 
-KeyTreeNode *BuildKeyTree(ActionCodeRec *ac_table)
+KeyTreeNode *BuildKeyTree(const ActionCodeRec *ac_table)
 {
    KeyTreeNode *top=0;
    char  term_name[256];
@@ -539,7 +554,7 @@ KeyTreeNode *BuildKeyTree(ActionCodeRec *ac_table)
       {
 	 KeyTreeNode *curr=top;
 
-	 char *code=ac_table->code;
+	 const char *code=ac_table->code;
 	 int delay=MAX_DELAY;
 
 	 fk_num=0;
@@ -779,26 +794,24 @@ void  ReadActionMap(FILE *f)
    NewTable[CurrTableCell].code=NULL;
 
    ActionCodeTable=NewTable;
-   NeedFreeActionCodeTable=true;
+   DynamicActionCodeTable=NewTable;
 }
 
 void FreeActionCodeTable()
 {
-   if(!NeedFreeActionCodeTable)
+   if(DynamicActionCodeTable)
    {
-      ActionCodeTable=0;
-      return;
+      for(int i=0; DynamicActionCodeTable[i].code; i++)
+	 free(DynamicActionCodeTable[i].code);
+      free(DynamicActionCodeTable);
+      DynamicActionCodeTable=0;
    }
-   for(int i=0; ActionCodeTable[i].code; i++)
-      free(ActionCodeTable[i].code);
-   free(ActionCodeTable);
    ActionCodeTable=0;
-   NeedFreeActionCodeTable=false;
 }
 
 int   GetNextAction()
 {
-   char  *store;
+   unsigned char *store;
    int   key;
    bool	 seen_func_key=false;
 
