@@ -16,7 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: options.cc,v 1.26 2001/10/03 06:25:10 lav Exp $ */
+/* $Id: options.cc,v 1.34 2006/11/29 07:36:35 lav Exp $ */
 
 #include <config.h>
 #include <ctype.h>
@@ -35,6 +35,7 @@
 #include "format.h"
 #include "colormnu.h"
 #include "mouse.h"
+#include "undo.h"
 
 bool ExplicitInitName=false;
 char InitName[256];
@@ -44,6 +45,10 @@ extern int MaxBackup;
 extern int le_use_default_colors;
 
 int useidl=0;
+int min_undo_levels=4;
+int max_undo_memory=128;
+int glue_small_changes=1;
+int undo_enable=1;
 
 extern int grsetno;
 
@@ -62,7 +67,7 @@ static int OldTabSize;
 
 struct  opt
 {
-   char  *name;
+   const char *name;
    int    type;
    void  *var;
    int    x,y;
@@ -83,13 +88,14 @@ struct  opt
 
 // {"rectangle &Blocks",ONE,  (void*)&rblock,  45,2},
 {"No regular expr.", ONE,  (void*)&noreg,	      45,2},
-{"Use &colors",      ONE,  (void*)&UseColor,	      45,3},
-{"&Syntax highlight",ONE,  (void*)&hl_option,	      45,4},
-{"Use tabs",         ONE,  (void*)&UseTabs,	      45,5},
-{"BackSp unindents", ONE,  (void*)&BackspaceUnindents,45,6},
-{"Lazy page scroll", ONE,  (void*)&PreferPageTop,     45,7},
+{"Match case",	     ONE,  (void*)&match_case,	      45,3},
+{"Use &colors",      ONE,  (void*)&UseColor,	      45,4},
+{"&Syntax highlight",ONE,  (void*)&hl_option,	      45,5},
+{"Use tabs",         ONE,  (void*)&UseTabs,	      45,6},
+{"BackSp unindents", ONE,  (void*)&BackspaceUnindents,45,7},
+{"Lazy page scroll", ONE,  (void*)&PreferPageTop,     45,8},
 #ifdef WITH_MOUSE
-{"Use mouse",	     ONE,  (void*)&UseMouse,	      45,8},
+{"Use mouse",	     ONE,  (void*)&UseMouse,	      45,9},
 #endif
 
 {"&Latin",           MANY, (void*)&inputmode,	3,5},
@@ -165,6 +171,14 @@ struct  opt
 {"&Shell",           STR,  (void*)Shell,     _StrPos,5,_StrLen,sizeof(Shell)    },
 {"&Help",	     STR,  (void*)HelpCmd,   _StrPos,6,_StrLen,sizeof(HelpCmd)  },
 {NULL}
+},
+   UndoOptTbl[]=
+{
+{"Enable undo", ONE, &undo_enable, 3,2},
+{"Glue small changes together", ONE, &glue_small_changes, 3,3},
+{"Mininum undo levels", NUM, (void*)&min_undo_levels,     3,4,3},
+{"Maximum undo memory (Mb)", NUM, (void*)&max_undo_memory,3,5,4},
+{NULL}
 };
 
 const struct init init[]=
@@ -188,6 +202,7 @@ const struct init init[]=
    { "savepos",      NUM,  (void*)&SavePos            },
    { "savehst",      NUM,  (void*)&SaveHst            },
    { "noreg",        NUM,  (void*)&noreg              },
+   { "match_case",   NUM,  (void*)&match_case         },
    { "linelen",      NUM,  (void*)&LineLen            },
    { "leftmrg",      NUM,  (void*)&LeftMargin         },
    { "flnmarg",      NUM,  (void*)&FirstLineMargin    },
@@ -202,6 +217,11 @@ const struct init init[]=
    { "backupnum",    NUM,  &MaxBackup		      },
    { "preferpagetop",NUM,  &PreferPageTop	      },
    { "wordwrap",     NUM,  &wordwrap		      },
+   { "syntaxhl",     NUM,  &hl_option		      },
+   { "undo_enable",  NUM,  &undo_enable		      },
+   { "undo_min_levels",NUM,&min_undo_levels	      },
+   { "undo_max_memory",NUM,&max_undo_memory	      },
+   { "undo_glue",    NUM,  &glue_small_changes	      },
 #ifdef WITH_MOUSE
    { "usemouse",     NUM,  &UseMouse		      },
 #endif
@@ -289,7 +309,7 @@ void  SaveConfToFile(const char *f,const struct init *init)
    fclose(conf);
 }
 
-void  SaveConf(char *f)
+void  SaveConf(const char *f)
 {
    MessageSync("Saving the editor options...");
    SaveConfToFile(f,init);
@@ -512,6 +532,10 @@ void  CorrectParameters()
       MaxBackup=1;
    if(MaxBackup>99)
       MaxBackup=99;
+   if(min_undo_levels<1)
+      min_undo_levels=1;
+   if(max_undo_memory<1)
+      max_undo_memory=1;
 
    switch(ShowScrollBar)
    {
@@ -553,6 +577,10 @@ void  CorrectParameters()
    }
 
    idlok(stdscr,useidl);
+   undo.SetMinGroups(min_undo_levels);
+   undo.SetMaxSize(max_undo_memory*1024*1024);
+   undo.SetEnable(undo_enable);
+   undo.SetGlue(glue_small_changes);
 
    if(buffer)
    {
@@ -651,16 +679,17 @@ static bool InOptField(int y,int x,struct opt *o)
 
 void  W_Dialogue(struct opt *opt,
              const char *SetupHelp,const char *SetupTitle,
-             int (*EatKey)(int),int (*HandleButton)(char *,int))
+             int (*EatKey)(int),int (*HandleButton)(const char *,int))
 {
    int newitem=0;
    int first=1;
    char  s[512];
    struct opt *p,*p1,*n,*curr=opt;
-   int shift=0,pos=0,i,key=0,d,dist;
+   int shift=0,pos=0,col=0,i,key=0,d,dist;
    int   action=-1;
    int   OldShowStatusLine=ShowStatusLine;
-   attr  *a=Upper->a;
+   const attr *a=Upper->a;
+   int len,ch_len;
 
    OldTabSize=TabSize;
 
@@ -698,7 +727,7 @@ void  W_Dialogue(struct opt *opt,
       case(NUM):
          PutStr(p->x-1,p->y,"[");
 	 PutStr(p->x+(p->len?p->len:2),p->y,"]");
-         DisplayItem(p->x+4,p->y,p->name,DIALOGUE_WIN_ATTR);
+         DisplayItem(p->x+2+(p->len?p->len:2),p->y,p->name,DIALOGUE_WIN_ATTR);
          break;
       }
    }
@@ -724,7 +753,7 @@ void  W_Dialogue(struct opt *opt,
             break;
          case(STR):
             if(p==curr)
-               sprintf(s,"%-*.*s",p->len,p->len,(char*)(p->var)+shift);
+               sprintf(s,"%-*.*s",p->len,p->len,(char*)(p->var)+mb_get_pos_for_col((char*)(p->var),shift,strlen((char*)(p->var))));
             else
                sprintf(s,"%-*.*s",p->len,p->len,(char*)(p->var));
             if(curr==p)
@@ -757,7 +786,7 @@ void  W_Dialogue(struct opt *opt,
             break;
          case(STR):
             curs_set(1);
-            GotoXY(curr->x+pos-shift,curr->y);
+            GotoXY(curr->x+col-shift,curr->y);
             break;
          case(NUM):
             curs_set(1);
@@ -884,12 +913,14 @@ use_key:
                if(p1->type==BUTTON)
                {
                   action=NEWLINE;
-                  strcpy(StringTyped,"\n");
+                  StringTyped[0]='\n';
+		  StringTypedLen=1;
                }
                else if(p1->type==ONE || p1->type==MANY)
                {
                   action=NO_ACTION;
-                  strcpy(StringTyped," ");
+                  StringTyped[0]=' ';
+		  StringTypedLen=1;
                }
                else
                   break;
@@ -902,6 +933,7 @@ use_key:
          newitem=0;
          shift=0;
          pos=0;
+	 col=0;
          first=1;
          continue;
       }
@@ -931,34 +963,37 @@ use_key:
          switch(action)
          {
          case(CHAR_LEFT):
-            if(pos>0)
-               pos--;
+	    if(pos>0)
+	       mb_char_left((char*)(curr->var),&pos,&col,strlen((char*)(curr->var)));
             break;
          case(CHAR_RIGHT):
             if(pos<(int)strlen((char*)(curr->var)))
-               pos++;
+	       mb_char_right((char*)(curr->var),&pos,&col,strlen((char*)(curr->var)));
             break;
          case(LINE_BEGIN):
-            pos=0;
+            pos=col=0;
             break;
          case(LINE_END):
             pos=strlen((char*)(curr->var));
+	    mb_get_col((char*)(curr->var),pos,&col,strlen((char*)(curr->var)));
             break;
          case(BACKSPACE_CHAR):
             if(pos==0)
                break;
+	    mb_char_left((char*)(curr->var),&pos,&col,strlen((char*)(curr->var)));
             if(shift>0)
                shift--;
-            pos--;
          case(DELETE_CHAR):
-            for(i=pos; ((char*)(curr->var))[i]; i++)
-               ((char*)(curr->var))[i]=((char*)(curr->var))[i+1];
+	    ch_len=mb_len((char*)(curr->var)+pos,strlen((char*)(curr->var))-pos);
+	    len=strlen((char*)(curr->var))-ch_len;
+            for(i=pos; i<=len; i++)
+               ((char*)(curr->var))[i]=((char*)(curr->var))[i+ch_len];
             break;
          case(DELETE_TO_EOL):
             ((char*)(curr->var))[pos]=0;
             break;
          case(ENTER_CHAR_CODE):
-            key=getcode();
+            key=getcode_char();
             goto do_insert;
          case(CHOOSE_CHAR):
             key=choose_ch();
@@ -985,12 +1020,13 @@ use_key:
             for(i=strlen((char*)(curr->var)); i>=pos; i--)
                ((char*)(curr->var))[i+1]=((char*)(curr->var))[i];
             ((char*)(curr->var))[pos++]=key;
+	    mb_get_col((char*)(curr->var),pos,&col,strlen((char*)(curr->var)));
          }
          first=0;
-         if(pos-shift>=curr->len)
-            shift=pos-curr->len+1;
-         if(pos-shift<0)
-            shift=pos;
+         if(col-shift>=curr->len)
+            shift=col-curr->len+1;
+         if(col-shift<0)
+            shift=col;
       }
    }
 leave_cycle:
@@ -1030,9 +1066,9 @@ leave_cycle:
    idlok(stdscr,useidl);
 }
 
-void  Dialogue(struct opt *opt,int WinWidth,int WinHeight,char *WinTitle,
+void  Dialogue(struct opt *opt,int WinWidth,int WinHeight,const char *WinTitle,
              const char *SetupHelp,const char *SetupTitle,
-             int (*EatKey)(int),int (*HandleButton)(char *,int))
+             int (*EatKey)(int),int (*HandleButton)(const char *,int))
 {
    WIN *optw;
    optw=CreateWin(MIDDLE,MIDDLE,WinWidth,WinHeight,DIALOGUE_WIN_ATTR,WinTitle,0);
@@ -1059,7 +1095,7 @@ int    OptEatKey(int k)
    return(-1);
 }
 
-int    OptHandleBut(char *,int)
+int    OptHandleBut(const char *,int)
 {
    return(0);
 }
@@ -1086,7 +1122,7 @@ int   TOEatKey(int k)
    (void)k;
    return(-1);
 }
-int   TOHandleBut(char *,int)
+int   TOHandleBut(const char *,int)
 {
    return(0);
 }
@@ -1107,6 +1143,10 @@ void  ProgOpt(void)
 {
    Dialogue(ProgOptTbl,70,9," External Programs ",NULL,NULL,TOEatKey,TOHandleBut);
 }
+void  UndoOpt(void)
+{
+   Dialogue(UndoOptTbl,40,8," Undo Options ",NULL,NULL,TOEatKey,TOHandleBut);
+}
 
 static int bg,fg,c_bold,c_rev,c_ul,c_dim,b_bold,b_rev,b_ul,b_dim;
 
@@ -1186,7 +1226,7 @@ static color new_color_pal[MAX_COLOR_NO+1];
 static color new_bw_pal[MAX_COLOR_NO+1];
 static bool color_applied;
 
-int ColorHandleBut(char *button,int index)
+int ColorHandleBut(const char *button,int index)
 {
    static int color_xlat[]={
       NORMAL_TEXT,BLOCK_TEXT,STATUS_LINE,SCROLL_BAR,ERROR_WIN,VERIFY_WIN,
@@ -1255,6 +1295,5 @@ void  ColorsOpt()
       clearok(stdscr,1);
       flag=REDISPLAY_ALL;
       RedisplayAll();
-      StatusLine();
    }
 }
