@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1993-2004 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -16,12 +16,13 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-/* $Id: textpoin.cc,v 1.11 2001/04/13 08:56:02 lav Exp $ */
+/* $Id: textpoin.cc,v 1.18 2006/06/20 12:03:03 lav Exp $ */
 
 #include <config.h>
 #include <limits.h>
 #include <stdlib.h>
 #include "edit.h"
+#include "mb.h"
 
 TextPoint   CurrentPos;
 TextPoint   ScreenTop;
@@ -215,7 +216,7 @@ void  TextPoint::FindOffset()
    }
    while(c>col)
    {
-      if(CharAt(o-1)=='\t')
+      if(CharAt(o-1)=='\t' || mb_mode)
       {
          o=LineBegin(o);
          c=0;
@@ -231,8 +232,8 @@ void  TextPoint::FindOffset()
       if(CharAt(o)=='\t')
          c=Tabulate(c);
       else
-         c++;
-      o++;
+         c+=CharWidthAt(o);
+      o+=CharSizeAt(o);
    }
    col=c;
    offset=o;
@@ -260,6 +261,8 @@ void  TextPoint::FindLineCol()
 	 {
 	    dist=abs(this->offset-scan->offset);
 	    found=scan;
+	    if(dist==0)
+	       break;
 	 }
       }
    }
@@ -288,8 +291,11 @@ void  TextPoint::FindLineCol()
    {
       if(BolAt(o) || CharAt(o-1)=='\t')
          break;
-      c--;
-      o--;
+      MBCheckLeftAt(o);
+      c-=MBCharWidth;
+      if(c<0)  // this would indicate libc bug with mb chars.
+	 break;
+      o-=MBCharSize;
    }
    if(o>offset)
    {
@@ -301,18 +307,45 @@ void  TextPoint::FindLineCol()
          l--;
       }
    }
+   while(o<offset)
+   {
+      offs next_line=NextLine(o);
+      if(next_line>offset || !BolAt(next_line))
+	 break;
+      o=next_line;
+      l++;
+      c=0;
+   }
    while(o<offset && !EofAt(o))
    {
       if(BolAt(o+1))
       {
          l++;
          c=0;
+	 o++;
+	 continue;
       }
       else if(CharAt(o)=='\t')
+      {
          c=Tabulate(c);
+	 o++;
+      }
+      else if(MBCheckLeftAt(o+1))
+      {
+         c+=MBCharWidth;
+	 o++;
+      }
       else
-         c++;
-      o++;
+      {
+	 MBCheckAt(o);
+         if(o+MBCharSize>offset)
+	 {
+	    o=offset;
+	    break;
+	 }
+	 c+=MBCharWidth;
+	 o+=MBCharSize;
+      }
    }
    col=c;
    line=l;
