@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-1997 by Alexander V. Lukyanov (lav@yars.free.net)
+ * Copyright (c) 1993-2005 by Alexander V. Lukyanov (lav@yars.free.net)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,8 +20,9 @@
 #include <string.h>
 #include <stdlib.h>
 #include "edit.h"
+#include "xalloca.h"
 
-attr  *curr_attr;
+const attr *curr_attr;
 
 WIN      *Upper = NULL;
 
@@ -56,19 +57,19 @@ void  Clear()
          mvaddch(y+Upper->y,x+Upper->x,' ');
    if(Upper->w>2 && Upper->h>2)
    {
-      PutCh(0,0,ACS_ULCORNER);
-      PutCh(Upper->w-1,0,ACS_URCORNER);
-      PutCh(0,Upper->h-1,ACS_LLCORNER);
-      PutCh(Upper->w-1,Upper->h-1,ACS_LRCORNER);
+      PutACS(0,0,ULCORNER);
+      PutACS(Upper->w-1,0,URCORNER);
+      PutACS(0,Upper->h-1,LLCORNER);
+      PutACS(Upper->w-1,Upper->h-1,LRCORNER);
       for(x=1; x<Upper->w-1; x++)
       {
-         PutCh(x,0,ACS_HLINE);
-         PutCh(x,Upper->h-1,ACS_HLINE);
+         PutACS(x,0,HLINE);
+         PutACS(x,Upper->h-1,HLINE);
       }
       for(y=1; y<Upper->h-1; y++)
       {
-         PutCh(0,y,ACS_VLINE);
-         PutCh(Upper->w-1,y,ACS_VLINE);
+         PutACS(0,y,VLINE);
+         PutACS(Upper->w-1,y,VLINE);
       }
       PutStr(MIDDLE,0,Upper->title);
    }
@@ -83,6 +84,46 @@ void  PutCh(int x,int y,chtype ch)
       addch_visual(ch);
    }
 }
+#if USE_MULTIBYTE_CHARS
+void  PutWCh(int x,int y,wchar_t ch)
+{
+   Absolute(&x,1,Upper->w);
+   Absolute(&y,1,Upper->h);
+   if(x>=0 && y>=0 && x<Upper->w && y<Upper->h)
+   {
+      move(y+Upper->y,x+Upper->x);
+      wchar_t vch=visualize_wchar(ch);
+      if(vch!=ch)
+	 attrset(curr_attr->so_attr);
+      int w=wcwidth(vch);
+      if(w==0)
+      {
+	 wchar_t a[2];
+	 a[0]=' ';
+	 a[1]=vch;
+	 addnwstr(a,2);
+      }
+      else if(w>0)
+	 addnwstr(&vch,1);
+      else if(ch<256)
+	 addch_visual(ch&255);
+      else
+         addch_visual('?');
+      attrset(curr_attr->n_attr);
+   }
+}
+void  PutCCh(int x,int y,cchar_t *ch)
+{
+   Absolute(&x,1,Upper->w);
+   Absolute(&y,1,Upper->h);
+   if(x>=0 && y>=0 && x<Upper->w && y<Upper->h)
+   {
+      move(y+Upper->y,x+Upper->x);
+      add_wch(ch);
+      attrset(curr_attr->n_attr);
+   }
+}
+#endif
 void  GotoXY(int x,int y)
 {
    Absolute(&x,1,Upper->w);
@@ -99,39 +140,89 @@ void  PutStr(int x,int y,const char *str)
 
    int   bx=x;
 
-   while(*str && y<Upper->h)
+#if USE_MULTIBYTE_CHARS
+   if(mb_mode)
    {
-      if(*str=='\n')
+      int len=strlen(str);
+      wchar_t *wstr=(wchar_t*)alloca((len+1)*sizeof(wchar_t));
+      memset(wstr,0,(len+1)*sizeof(wchar_t));
+      mbstowcs(wstr,str,len);
+
+      while(*wstr && y<Upper->h)
       {
-         while(x<Upper->w-1)
-            mvaddch(y+Upper->y,(x++)+Upper->x,' ');
-         x=bx;
-         y++;
-      }
-      else if(*str=='\t')
-      {
-	 int add=((x-bx+8)&~7)-x+bx;
-	 while(add-->0)
+	 attrset(curr_attr->n_attr);
+	 if(*wstr=='\n')
 	 {
-	    if(x<Upper->w)
-	       mvaddch(y+Upper->y,x+Upper->x,' ');
+	    while(x<Upper->w-1)
+	       mvaddch(y+Upper->y,(x++)+Upper->x,' ');
+	    x=bx;
+	    y++;
+	 }
+	 else if(*wstr=='\t')
+	 {
+	    int add=((x-bx+8)&~7)-x+bx;
+	    while(add-->0)
+	    {
+	       if(x<Upper->w)
+		  mvaddch(y+Upper->y,x+Upper->x,' ');
+	       x++;
+	    }
+	 }
+	 else
+	 {
+	    wchar_t wc=visualize_wchar(*wstr);
+	    int width=wcwidth(wc);
+	    if(x>=0 && y>=0 && x+width<=Upper->w)
+	    {
+	       move(y+Upper->y,x+Upper->x);
+	       if(wc!=*wstr)
+		  attrset(curr_attr->so_attr);
+	       addnwstr(&wc,1);
+	       x=getcurx(stdscr)-Upper->x;
+	    }
+	    else
+	       x+=width;
+	 }
+	 wstr++;
+      }
+   }
+   else	 // note the following block
+#endif // !USE_MULTIBYTE_CHARS
+   {
+      while(*str && y<Upper->h)
+      {
+	 if(*str=='\n')
+	 {
+	    while(x<Upper->w-1)
+	       mvaddch(y+Upper->y,(x++)+Upper->x,' ');
+	    x=bx;
+	    y++;
+	 }
+	 else if(*str=='\t')
+	 {
+	    int add=((x-bx+8)&~7)-x+bx;
+	    while(add-->0)
+	    {
+	       if(x<Upper->w)
+		  mvaddch(y+Upper->y,x+Upper->x,' ');
+	       x++;
+	    }
+	 }
+	 else
+	 {
+	    if(x>=0 && y>=0 && x<Upper->w)
+	    {
+	       move(y+Upper->y,x+Upper->x);
+	       addch_visual((byte)*str);
+	    }
 	    x++;
 	 }
+	 str++;
       }
-      else
-      {
-         if(x>=0 && y>=0 && x<Upper->w)
-	 {
-            move(y+Upper->y,x+Upper->x);
-	    addch_visual((byte)*str);
-	 }
-	 x++;
-      }
-      str++;
    }
 }
 
-WIN   *CreateWin(int x,int y,unsigned w,unsigned h,struct attr *a,
+WIN   *CreateWin(int x,int y,unsigned w,unsigned h,const attr *a,
                  const char *title,int flags)
 {
    WIN   *win;
@@ -171,27 +262,29 @@ void  DestroyWin(WIN *win)
 }
 void  DisplayWin(WIN *win)
 {
-   chtype   *save;
+   win_cell *save;
    int      x,y;
 
    curs_set(0);
+   attrset(0);
 
    win->prev=Upper;
    Upper=win;
 
-   save = win->buf = (chtype*)malloc((win->h+1)*(win->w+2)*sizeof(chtype));
+   save = win->buf = (win_cell*)malloc((win->h+1)*(win->w+2)*sizeof(win_cell));
 
    for(y=0; y<win->h+1; y++)
    {
       for(x=0; x<win->w+2; x++)
       {
-         *save=mvinch(y+win->y,x+win->x);
+         scr_get_cell(y+win->y,x+win->x,save);
          if(!(y<win->h && x<win->w) && !(win->flags&NOSHADOW) && x>1 && y>0)
          {
             if(y+win->y<LINES && x+win->x<COLS)
             {
-               attrset(SHADOW_ATTR->n_attr);
-               mvaddch(y+win->y,x+win->x,*save&(A_CHARTEXT|A_ALTCHARSET));
+	       win_cell ch=*save;
+	       win_cell_set_attrs(&ch,SHADOW_ATTR->n_attr);
+               scr_put_cell(y+win->y,x+win->x,&ch);
             }
          }
          save++;
@@ -203,7 +296,7 @@ void  DisplayWin(WIN *win)
 }
 void  CloseWin()
 {
-   chtype   *save;
+   win_cell *save;
    int      x,y;
 
    save = Upper->buf;
@@ -215,7 +308,7 @@ void  CloseWin()
       {
 	 if(y+Upper->y<LINES && x+Upper->x<COLS
 	 && x>=0 && y>=0 && (x<Upper->w || !(Upper->flags&NOSHADOW)))
-   	    mvaddch(y+Upper->y,x+Upper->x,*save);
+   	    scr_put_cell(y+Upper->y,x+Upper->x,save);
       }
 
    free(Upper->buf);
